﻿/******************************************************************************************************************************************
 *
 * Filename    : Program.cs
 * Project     : VectorReflection
 * Description : This is an visualization how a vector is reflected at a plane. This visualiztion is in 2D, but the used mathematics
 *               also applies for higher dimensions.
 *
 * Sunshine, March 2k21
 * www.sunshine2k.de | www.bastian-molkenthin.de
 *
 * This file is licensed under MIT licence. For further details, see http://www.sunshine2k.de/license.html.
 *
 * Update October 2k23: Fix issue with two calls of requestAnimationFrame()
 *****************************************************************************************************************************************/

let cvs: HTMLCanvasElement;
let myApp: VectorReflectionApp;

class VectorReflectionApp
{
  private canvasHtmlElement: HTMLCanvasElement;
  private cvsContext: CanvasRenderingContext2D;

  private planePoint1: MovablePoint;
  private planePoint2: MovablePoint;
  private sourceVecPoint1: MovablePoint;  // second source vector point is always midpoint of plane
  private points: Array<MovablePoint>;

  private vPlaneMidpoint: Vector2D;
  private vPlane: Vector2D;
  private vPlaneNormal: Vector2D;
  private vSource: Vector2D;
  private vReflection: Vector2D;

  constructor(cvs: HTMLCanvasElement)
  {
    this.canvasHtmlElement = cvs;
    this.cvsContext = <CanvasRenderingContext2D>this.canvasHtmlElement.getContext("2d");

    this.planePoint1 = new MovablePoint(this.cvsContext, 100, 300, "");
    this.planePoint2 = new MovablePoint(this.cvsContext, 400, 200, "");

    this.sourceVecPoint1 = new MovablePoint(this.cvsContext, 120, 180, "");

    this.points = new Array<MovablePoint>();
    this.points.push(this.planePoint1);
    this.points.push(this.planePoint2);
    this.points.push(this.sourceVecPoint1);

    this.canvasHtmlElement.addEventListener("mousedown", this.mouseDown, false);
    this.canvasHtmlElement.addEventListener("mouseup", this.mouseUp, false);
    this.canvasHtmlElement.addEventListener("mousemove", this.mouseMove, false);

    this.vReflection = new Vector2D(0, 0);

    this.update(0);
  }

  /**
   * Update loop that does cyclically all calculation and drawing.
   */
  private update = (time: number): void =>
  {
    this.calculate();

    this.paint(time);

    window.requestAnimationFrame(this.update);
  }

  /**
   * Performs the actual vector calculation
   */
  private calculate = () : void =>
  {
    /* calculate plane vector and plane normal (normalized!) */
    this.vPlane = new Vector2D(this.planePoint2.getX() - this.planePoint1.getX(), this.planePoint2.getY() - this.planePoint1.getY());
    this.vPlaneNormal = this.vPlane.getPerpendicularCW().normalize();

    /* calculate plane midpoint */
    this.vPlaneMidpoint = new Vector2D(
      this.planePoint1.getX() + (this.planePoint2.getX() - this.planePoint1.getX()) * 0.5,
      this.planePoint1.getY() + (this.planePoint2.getY() - this.planePoint1.getY()) * 0.5);

    /* calculate source vector (to be projected) from its points */
    this.vSource = new Vector2D(this.vPlaneMidpoint.X - this.sourceVecPoint1.getX(), this.vPlaneMidpoint.Y - this.sourceVecPoint1.getY());

    /* calculate dot product of source vector and plane normal vector */
    let dotSourceNormal: number = this.vSource.dotProduct(this.vPlaneNormal);

    /* calculate actual reflected vector */
    this.vReflection = this.vSource.getSubVec(this.vPlaneNormal.getMul(2.0 * dotSourceNormal));

  }

  /**
   * Performs all drawing
   */
  private paint = (time: number): void =>
  {
    this.cvsContext.save();

    /* fill background */
    this.cvsContext.fillStyle = "#ffdda8";
    this.cvsContext.fillRect(0, 0, this.canvasHtmlElement.width, this.canvasHtmlElement.height);

    /* draw plane line */
    this.cvsContext.strokeStyle = "#646464";
    this.drawLine(this.planePoint1.getX(), this.planePoint1.getY(), this.planePoint2.getX(), this.planePoint2.getY());

    /* draw plane normal in the middle of the plane */
    this.cvsContext.strokeStyle = "#CCCCCC";
    this.drawLine(this.vPlaneMidpoint.X, this.vPlaneMidpoint.Y,
                  this.vPlaneMidpoint.X + (this.vPlaneNormal.X * 100), this.vPlaneMidpoint.Y + (this.vPlaneNormal.Y * 100));

    /* draw input vector */
    let sourceVector: Vector2D = new Vector2D(this.vPlaneMidpoint.X - this.sourceVecPoint1.getX(), this.vPlaneMidpoint.Y - this.sourceVecPoint1.getY());
    this.cvsContext.strokeStyle = "#6de900";
    this.drawLine(this.vPlaneMidpoint.X, this.vPlaneMidpoint.Y, this.vPlaneMidpoint.X - sourceVector.X, this.vPlaneMidpoint.Y - sourceVector.Y);

    /* draw reflected vector */
    let vReflectionNormal = this.vReflection.getNormalized().mul(this.vSource.getLength());
    this.cvsContext.strokeStyle = "#03c6fc";
    this.drawLine(this.vPlaneMidpoint.X, this.vPlaneMidpoint.Y, this.vPlaneMidpoint.X + vReflectionNormal.X, this.vPlaneMidpoint.Y + vReflectionNormal.Y);

    /* draw points */
    this.cvsContext.strokeStyle = "#000000";
    this.cvsContext.lineWidth = 0.7;
    for (let i = 0; i < this.points.length; i++) 
    {
      this.points[i].draw();
    }

    this.cvsContext.restore();
  }

  private drawLine = (x1: number, y1: number, x2: number, y2: number): void =>
  {
    this.cvsContext.beginPath();
    this.cvsContext.lineWidth = 1;
    this.cvsContext.moveTo(x1, y1);
    this.cvsContext.lineTo(x2, y2);
    this.cvsContext.stroke();
  }

  public mouseDown = (event: MouseEvent): void =>
  {
    let cursorPosX: number = event.x - this.canvasHtmlElement.offsetLeft;
    let cursorPosY: number = event.y - this.canvasHtmlElement.offsetTop;

    for (let i = 0; i < this.points.length; i++) 
    {
      if (this.points[i].startDragging(cursorPosX, cursorPosY)) 
      {
        this.canvasHtmlElement.style.cursor = "pointer";
        break;
      }
    }
  }

  public mouseUp = (event: MouseEvent): void => 
  {
    for (let i = 0; i < this.points.length; i++) 
    {
      this.points[i].stopDragging();
    }

    /* change cursor back to normal */
    this.canvasHtmlElement.style.cursor = "default";
  }

  public mouseMove = (event: MouseEvent): void => 
  {
    var cursorPosX: number = event.x - this.canvasHtmlElement.offsetLeft;
    var cursorPosY: number = event.y - this.canvasHtmlElement.offsetTop;

    var pointHit: boolean = false;

    /* inform each point about new cursor position */
    for (let i = 0; i < this.points.length; i++) 
    {
      this.points[i].cursorMoved(cursorPosX, cursorPosY);
      /* if cursor is over point, change cursor to hand */
      if (this.points[i].isCursorOnPoint(cursorPosX, cursorPosY)) 
      {
        this.canvasHtmlElement.style.cursor = "pointer";
        pointHit = true;
      }
    }

    /* if cursor is over no point, use normal cursor */
    if (!pointHit)
    {
      this.canvasHtmlElement.style.cursor = "default";
    }
  }
}


window.addEventListener("load", () =>
{
  cvs = <HTMLCanvasElement>document.getElementById('cvs');
  myApp = new VectorReflectionApp(cvs);
});